/* MiniDLNA media server
 * Copyright (C) 2009-2010  Justin Maggard
 *
 * This file is part of MiniDLNA.
 *
 * MiniDLNA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * MiniDLNA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MiniDLNA. If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <libgen.h>
#include <limits.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "tagutils/tagutils.h"

#include "upnpglobalvars.h"
#include "scanner.h"
#include "metadata.h"
#include "utils.h"
#include "sql.h"
#include "log.h"

#if 1 // support TRANSCODE : support streaming & DVDISO
#include <signal.h>
#include <sys/wait.h>
#include <poll.h>
#include <errno.h>
#include "process.h"
#ifdef __CYGWIN__
#include <sys/cygwin.h>
#endif //__CYGWIN__

void
insert_containers(const char *, const char *, const char *, const char *, int64_t );

#define R (0)
#define W (1)

#define CANNOTEXECMPLAYER 10

static int
exec_mplayer(char *title, char *filename, int *fd_r, int *fd_w)
{
	char *args[] = {
#ifndef __CYGWIN__
		"mplayer",
#else //__CYGWIN__
		"./mplayer.exe",
#endif //__CYGWIN__
		title, "-dvd-device", filename,
		"-vo", "null", "-ao", "null", "-quiet", "-slave", "-endpos", "0",
		NULL
	};

	int pipe_c2p[2], pipe_p2c[2];
	int pid, ret;

	/* Create a pipe. */
	if(pipe(pipe_c2p)<0) {
		perror("popen2");
		return -1;
	}
	if(pipe(pipe_p2c)<0){
		perror("popen2");
		close(pipe_c2p[R]);
		close(pipe_c2p[W]);
		return -1;
	}

	/* Invoke processs */
	if((pid=process_fork())<0){
		perror("can not fork to exec mplayer");
		close(pipe_c2p[R]);
		close(pipe_c2p[W]);
		close(pipe_p2c[R]);
		close(pipe_p2c[W]);
		return -1;
	}
	if(pid==0) {  /* I'm child */
		close(pipe_p2c[W]);
		close(pipe_c2p[R]);
		dup2(pipe_p2c[R],0);
		dup2(pipe_c2p[W],1);
		close(pipe_p2c[R]);
		close(pipe_c2p[W]);

		ret = execvp(args[0], args);

		if (ret < 0) {
			perror("can not exec mplayer");
			//close(pipe_p2c[R]);
			//close(pipe_c2p[W]);
		}
		exit(CANNOTEXECMPLAYER);
	}

	close(pipe_p2c[R]);
	close(pipe_c2p[W]);
	*fd_w=pipe_p2c[W];
	*fd_r=pipe_c2p[R];

	return pid;
}

static double
get_time_length(char *title, char *filename)
{
	#define BUFFSIZE 256
	char buf[BUFFSIZE];

	int pid, pid_status, ret, fd_r, fd_w, timeout;
	off_t read_stream_size=0;
	struct pollfd fds[2];
	char *cmd = "get_time_length\n get_audio_codec\n get_audio_samples\n get_video_resolution\n";
	int written=0;
	char *plength;
	float length=-1;

	static int mplayerExist=1;

	if (!mplayerExist)
		return -1;
	
	pid = exec_mplayer(title, filename, &fd_r, &fd_w);
	if (pid < 0)
		return -1;

	fds[0].fd = fd_r;
	fds[0].events =  POLLIN|POLLPRI;
	fds[0].revents = POLLIN|POLLPRI;
	fds[1].fd = fd_w;
	fds[1].events =  POLLOUT|POLLERR|POLLHUP|POLLNVAL;
	fds[1].revents = POLLOUT|POLLERR|POLLHUP|POLLNVAL;
	timeout = 10*1000; /* timeout = 10sec */

	while(1)
	{
		ret = waitpid(pid, &pid_status, WNOHANG | WUNTRACED | WCONTINUED); 
		if (WIFEXITED(pid_status) && (WEXITSTATUS(pid_status) == CANNOTEXECMPLAYER))
			mplayerExist = 0;
		else if (WIFEXITED(pid_status) && (ret == pid))
			break;
		if (ret == -1) {
			DPRINTF(E_WARN, L_SCANNER, "process PID(%d) : %s\n", (int)pid, strerror(errno));
			break;
		}
		ret = poll(fds, (nfds_t)2, timeout);
		if (ret <= 0) {
			DPRINTF(E_WARN, L_SCANNER, "poll error : No data in Pipe\n");
			break;
		}
		if (written == 0) {
		{
			if ((fds[1].revents & (POLLERR|POLLHUP|POLLNVAL)) == 0)
				ret = write(fd_w, cmd, strlen(cmd)); // write command to PIPE
				fds[1].events = 0;
				written = 1;
			}
		}
		if ((fds[0].revents & POLLIN) == 0)
			continue;
		read_stream_size = read(fd_r, buf, BUFFSIZE); // read status from PIPE
		if (read_stream_size == 0) {
			//DPRINTF(E_WARN, L_SCANNER, "reached to EOF in PID:%d\n", (int)getpid());
			break; //EOF
		}
		buf[read_stream_size] = '\0';
		//for (ret=0; ret<read_stream_size ; ret++) if (buf[ret] == '\r') buf[ret] = '\n';
		//DPRINTF(E_WARN, L_SCANNER, "%s", buf);
		plength = strstr(buf, "ANS_LENGTH=");
		if (plength) {
			length = atof(plength+11);
			DPRINTF(E_DEBUG, L_SCANNER, "%s", buf);
		}
	}

	//ret = waitpid(pid, &pid_status, 0); 
	ret = waitpid(pid, &pid_status, WNOHANG | WUNTRACED | WCONTINUED); 
	if ((ret != -1) && WIFEXITED(pid_status) && (WEXITSTATUS(pid_status) == CANNOTEXECMPLAYER))
		mplayerExist = 0;

	close(fd_r);
	close(fd_w);
	
	return length;
}

static int
insert_file_into_db_for_streaming(char *name, char *path, char *url, char *fullpath, int track)
{
	//name  : TITLE in DETAIL, NAME in OBJECTS
	//path  : PATH in DETAIL
	//url   : FILE=URL
	//track : FILE#

	char * parentID = NULL;
	char objectID[128], objectID2[128];
	static int64_t detailID;
	char *typedir_parentID;
	char *baseid;
	char base[8];
	int next_available_id;
	char *mime="audio/mpeg";
	char *artist="NETRADIO";
	char *class = "item.audioItem.musicTrack";
	char /*fullpath[PATH_MAX],*/ *pfullpath;
	char title[256];
	char duration[14];
	
	if (strlen(url) >= 512)
		return 0;

	if ( (track > 1) && (strncmp(url, "DVDISO-", 7)) )
	{
		char *newUrl, *curUrl;
		if (track > 8)
			return 0;
		curUrl  = sql_get_text_field(db, "SELECT COMMENT from DETAILS where ID = '%lld'", detailID);
		if (curUrl == NULL)
			return 0;
		xasprintf(&newUrl, "%s$SEPARATOR$%s", curUrl, url);
		sqlite3_free(curUrl);
		if (strlen(newUrl) <512)
			sql_exec(db, "UPDATE DETAILS set COMMENT = %Q where ID = %lld", newUrl, detailID);
		free(newUrl);
		return 0;
	}

	strcpy(base, MUSIC_DIR_ID);

	if ( !strncmp(url, "NETTV-", 6) || !strncmp(url, "DVDISO-", 7))
	{
		mime =   "video/mpeg";
		artist = "NETTV";
		class =  "item.videoItem";
		strcpy(base, VIDEO_DIR_ID);
	}

	if (strncmp(url, "DVDISO-", 7))
	{
		strcpy(duration, "24:00:00.000");
		strncpy(title, name, sizeof(title));
		snprintf(fullpath, PATH_MAX, "%s/%s.dmy", path, name);
	}
	else
	{
		double lengthd;
		long  lengthl;
		char *trackB, *trackE;

		pfullpath = strstr(url, "file://")+7;
		strncpy(fullpath, pfullpath, PATH_MAX);

#ifdef __CYGWIN__
		if (!cygwin_posix_path_list_p(pfullpath) || (strchr(fullpath, '\\') != NULL))
			cygwin_conv_path(CCP_WIN_A_TO_POSIX | CCP_ABSOLUTE, pfullpath, fullpath, PATH_MAX);
#endif // __CYGWIN__

		if (access(fullpath, R_OK) < 0)
		{
			DPRINTF(E_WARN, L_SCANNER, "file ""%s"" does not exist!\n", fullpath);
			return -1;
		}

		artist = "DVDISO";
		trackB = strstr(url, "dvd://")+6;
		trackE = strchr(trackB, ' ');
		*trackE = '\0';

		strncpy(title, strstr(url, "dvd://"), sizeof(title));
		lengthd = get_time_length(title, pfullpath);
		if (lengthd > 0)
		{
			lengthl = (long)(lengthd*1000);
			sprintf(duration, "%ld:%02ld:%02ld.%03ld",
						(lengthl/3600000%100),
						(lengthl/60000%60),
						(lengthl/1000%60),
						(lengthl%1000));
		}
		else
			strcpy(duration, "8:00:00.000"); // could not get duration

		snprintf(title, sizeof(title), "%s title#%s", name, trackB); 
		*trackE = ' ';

		snprintf(fullpath, PATH_MAX, "%s/%s.dmy", path, name);
	}

	detailID = sql_exec(db, "INSERT into DETAILS"
							" (PATH, SIZE, TIMESTAMP, DURATION, DATE, CHANNELS, BITRATE, SAMPLERATE, RESOLUTION,"
							"  TITLE, CREATOR, ARTIST, GENRE, COMMENT, DLNA_PN, MIME, ALBUM_ART, ALBUM) "
							"VALUES"
							" (%Q, %lld, %ld, %Q, %Q, %Q, %Q, %Q, %Q, '%q', %Q, %Q, %Q, %Q, %Q, '%q', %lld, %Q);",
							fullpath, (long long)100/*st_size*/, (long)0/*st_mtime*/, duration,
							""/*date*/, ""/*channels*/, NULL/*bitrate*/, NULL/*frequency*/, NULL/*resolution*/,
							title/*title*/, artist/*creator*/, artist, NULL/*genre*/, url/*comment*/, NULL/*dlna_pn*/,
							mime, (long long)0/*album_art*/, title/*album*/);
	if( detailID != SQLITE_OK )
	{
		DPRINTF(E_WARN, L_SCANNER, "Error inserting details for %s!\n", fullpath);
		return -1;
	}
	else
	{
		detailID = sqlite3_last_insert_rowid(db);
	}

	parentID = sql_get_text_field(db, "SELECT OBJECT_ID from OBJECTS o left join DETAILS d on (d.ID = o.DETAIL_ID)"
			                          " where d.PATH = '%q' and REF_ID is NULL", path);
	if (parentID)
	{
		next_available_id = get_next_available_id("OBJECTS", parentID);
		parentID += 2;
	}
	else
	{
		next_available_id = get_next_available_id("OBJECTS", BROWSEDIR_ID);
		parentID = "";
	}

	sprintf(objectID, "%s%s$%X", BROWSEDIR_ID, parentID, next_available_id);
	sql_exec(db, "INSERT into OBJECTS"
	             " (OBJECT_ID, PARENT_ID, CLASS, DETAIL_ID, NAME) "
	             "VALUES"
	             " ('%s', '%s%s', '%s', %lld, '%q')",
	             objectID, BROWSEDIR_ID, parentID, class, detailID, title);

	if( *parentID )
	{
		int typedir_objectID = 0;
		typedir_parentID = strdup(parentID);
		baseid = strrchr(typedir_parentID, '$');
		if( baseid )
		{
			typedir_objectID = strtol(baseid+1, NULL, 16);
			*baseid = '\0';
		}
		insert_directory(title, fullpath, base, typedir_parentID, typedir_objectID);
		free(typedir_parentID);
	}

	sprintf(objectID2, "%s%s$%X", base, parentID, next_available_id);
	sql_exec(db, "INSERT into OBJECTS"
	             " (OBJECT_ID, PARENT_ID, REF_ID, CLASS, DETAIL_ID, NAME) "
	             "VALUES"
	             " ('%s', '%s%s', '%s', '%s', %lld, '%q')",
	             objectID2, base, parentID, objectID, class, detailID, title);

	insert_containers(title, fullpath, objectID, class, detailID);

	if ( *parentID )
		sqlite3_free(parentID-2);

	return 1;
}
#endif // support TRANSCODE : support streaming & DVDISO

int
insert_playlist(const char * path, char * name)
{
	struct song_metadata plist;
	struct stat file;
	int items = 0, matches, ret;
	char type[4];
#if 1 // support TRANSCODE : support streaming
	int steaming = 0;
#endif // support TRANSCODE : support streaming

	strncpy(type, strrchr(name, '.')+1, 4);

	if( start_plist(path, NULL, &file, NULL, type) != 0 )
	{
		DPRINTF(E_WARN, L_SCANNER, "Bad playlist [%s]\n", path);
		return -1;
	}
	while( (ret = next_plist_track(&plist, &file, NULL, type)) == 0 )
	{
		items++;
#if 1 // support TRANSCODE : support streaming
		if (strstr(plist.path, "://") && !strstr(plist.path, "dvd://"))
			steaming = 1;
#endif // support TRANSCODE : support streaming
		freetags(&plist);
	}
	if( ret == 2 ) // Bad playlist -- contains binary characters
	{
		DPRINTF(E_WARN, L_SCANNER, "Bad playlist [%s]\n", path);
		return -1;
	}
	strip_ext(name);

	DPRINTF(E_DEBUG, L_SCANNER, "Playlist %s contains %d items\n", name, items);

#if 1 // support TRANSCODE : support streaming
	if (steaming)
		items = 1;
#endif // support TRANSCODE

	matches = sql_get_int_field(db, "SELECT count(*) from PLAYLISTS where NAME = '%q'", name);
	if( matches > 0 )
	{
		sql_exec(db, "INSERT into PLAYLISTS"
		             " (NAME, PATH, ITEMS) "
	        	     "VALUES"
		             " ('%q(%d)', '%q', %d)",
		             name, matches, path, items);
	}
	else
	{
		sql_exec(db, "INSERT into PLAYLISTS"
		             " (NAME, PATH, ITEMS) "
	        	     "VALUES"
		             " ('%q', '%q', %d)",
		             name, path, items);
	}
	return 0;
}

static unsigned int
gen_dir_hash(const char *path)
{
	char dir[PATH_MAX], *base;
	int len;

	strncpy(dir, path, sizeof(dir));
	dir[sizeof(dir)-1] = '\0';
	base = strrchr(dir, '/');
	if( !base )
		base = strrchr(dir, '\\');
	if( base )
	{
		*base = '\0';
		len = base - dir;
	}
	else
		return 0;
	

	return DJBHash(dir, len);
}

int
fill_playlists()
{
	int rows, i, found, len;
	char **result;
	char *plpath, *plname, *fname, *last_dir;
	unsigned int hash, last_hash = 0;
	char class[] = "playlistContainer";
	struct song_metadata plist;
	struct stat file;
	char type[4];
	int64_t plID, detailID;
	char sql_buf[] = "SELECT ID, NAME, PATH from PLAYLISTS where ITEMS > FOUND";

	DPRINTF(E_WARN, L_SCANNER, "Parsing playlists...\n");

	if( sql_get_table(db, sql_buf, &result, &rows, NULL) != SQLITE_OK ) 
		return -1;
	if( !rows )
		goto done;

	rows++;
	for( i=3; i<rows*3; i++ )
	{
		plID = strtoll(result[i], NULL, 10);
		plname = result[++i];
		plpath = result[++i];
		last_dir = NULL;
		last_hash = 0;

		strncpy(type, strrchr(plpath, '.')+1, 4);

		if( start_plist(plpath, NULL, &file, NULL, type) != 0 )
			continue;

		DPRINTF(E_DEBUG, L_SCANNER, "Scanning playlist \"%s\" [%s]\n", plname, plpath);
		if( sql_get_int_field(db, "SELECT ID from OBJECTS where PARENT_ID = '"MUSIC_PLIST_ID"'"
		                          " and NAME = '%q'", plname) <= 0 )
		{
			detailID = GetFolderMetadata(plname, NULL, NULL, NULL, 0);
			sql_exec(db, "INSERT into OBJECTS"
			             " (OBJECT_ID, PARENT_ID, DETAIL_ID, CLASS, NAME) "
			             "VALUES"
			             " ('%s$%llX', '%s', %lld, 'container.%s', '%q')",
			             MUSIC_PLIST_ID, plID, MUSIC_PLIST_ID, detailID, class, plname);
		}

		plpath = dirname(plpath);
		found = 0;
		while( next_plist_track(&plist, &file, NULL, type) == 0 )
		{
#if 1 // support TRANSCODE : support streaming & DVDISO
			char fullpath[PATH_MAX];
			int streaming = 0;
			if (strstr(plist.path, "://"))
			{
				found++;
				if (insert_file_into_db_for_streaming(plname, plpath, plist.path, fullpath, found) != 1)
				{
					freetags(&plist);
					continue;
				}
				streaming = 1;
			}
#endif // support TRANSCODE : support streaming & DVDISO
			hash = gen_dir_hash(plist.path);
			if( sql_get_int_field(db, "SELECT 1 from OBJECTS where OBJECT_ID = '%s$%llX$%d'",
			                      MUSIC_PLIST_ID, plID, plist.track) == 1 )
			{
				//DEBUG DPRINTF(E_DEBUG, L_SCANNER, "%d: already in database\n", plist.track);
				found++;
       				freetags(&plist);
				continue;
			}
			if( last_dir )
			{
				if( hash == last_hash )
				{
					fname = basename(plist.path);
#if 1 // support TRANSCODE : support streaming & DVDISO
					if (streaming)
						fname = basename(fullpath);
#endif // support TRANSCODE : support streaming & DVDISO
					detailID = sql_get_int_field(db, "SELECT ID from DETAILS where PATH = '%q/%q'", last_dir, fname);
				}
				else
					detailID = -1;
				if( detailID <= 0 )
				{
					sqlite3_free(last_dir);
					last_dir = NULL;
				}
				else
					goto found;
			}

#if 0			
			fname = plist.path;
			DPRINTF(E_DEBUG, L_SCANNER, "%d: checking database for %s\n", plist.track, plist.path);
#else // support TRANSCODE : support streaming & DVDISO
			if (!streaming)
				fname = plist.path;
			else
				fname = fullpath;
			DPRINTF(E_DEBUG, L_SCANNER, "%d: checking database for %s\n", plist.track, fname);
#endif // support TRANSCODE : support streaming & DVDISO
			if( !strpbrk(fname, "\\/") )
			{
				len = strlen(fname) + strlen(plpath) + 2;
				plist.path = malloc(len);
				snprintf(plist.path, len, "%s/%s", plpath, fname);
				free(fname);
				fname = plist.path;
			}
			else
			{
				while( *fname == '\\' )
				{
					fname++;
				}
			}
retry:
			//DEBUG DPRINTF(E_DEBUG, L_SCANNER, "* Searching for %s in db\n", fname);
			detailID = sql_get_int_field(db, "SELECT ID from DETAILS where PATH like '%%%q'", fname);
			if( detailID > 0 )
			{
found:
				DPRINTF(E_DEBUG, L_SCANNER, "+ %s found in db\n", fname);
				sql_exec(db, "INSERT into OBJECTS"
				             " (OBJECT_ID, PARENT_ID, CLASS, DETAIL_ID, NAME, REF_ID) "
				             "SELECT"
				             " '%s$%llX$%d', '%s$%llX', CLASS, DETAIL_ID, NAME, OBJECT_ID from OBJECTS"
				             " where DETAIL_ID = %lld and OBJECT_ID glob '" BROWSEDIR_ID "$*'",
				             MUSIC_PLIST_ID, plID, plist.track,
				             MUSIC_PLIST_ID, plID,
				             detailID);
				if( !last_dir )
				{
					last_dir = sql_get_text_field(db, "SELECT PATH from DETAILS where ID = %lld", detailID);
					fname = strrchr(last_dir, '/');
					if( fname )
						*fname = '\0';
					last_hash = hash;
				}
				found++;
			}
			else
			{
				DPRINTF(E_DEBUG, L_SCANNER, "- %s not found in db\n", fname);
				if( strchr(fname, '\\') )
				{
					fname = modifyString(fname, "\\", "/");
					goto retry;
				}
				else if( (fname = strchr(fname, '/')) )
				{
					fname++;
					goto retry;
				}
			}
       			freetags(&plist);
		}
		if( last_dir )
		{
			sqlite3_free(last_dir);
			last_dir = NULL;
		}
		sql_exec(db, "UPDATE PLAYLISTS set FOUND = %d where ID = %lld", found, plID);
	}
done:
	sqlite3_free_table(result);
	DPRINTF(E_WARN, L_SCANNER, "Finished parsing playlists.\n");

	return 0;
}

